from collections import defaultdict
from typing import DefaultDict, Dict, List, Set, Tuple

import numpy as np
from recon.constants import NOT_LABELED
from recon.recognizer import EntityRecognizer
from recon.types import (
    Example,
    HardestExampleV2,
    LabelDisparity,
    PredictionError,
    PredictionErrorExamplePair,
)
from spacy.scorer import PRFScore
from wasabi import Printer


def get_ents_by_label(data: List[Example], case_sensitive: bool = False) -> DefaultDict[str, List[str]]:
    """Get a dictionary of unique text spans by label for your data

    # TODO: Ok so this needs to return more than just a set for each label.

    We want to return a dictionary that maps labels to AnnotationCount objects where each
    AnnotationCount contains the text of the annotation text, the total number of times it's mentioned (e.g. what entity_coverage does)
    but also the examples it is in.

    So maybe I can get this info from entity_coverage? IDK but this is dumb rn and not very flexible.

    Maybe I should keep this function returning a set of strings for each label for compatability but I need the other way too
    so I know what to focus on in editing and a

    Args:
        data (List[Example]): List of examples
        case_sensitive (bool, optional): Consider case of text for each annotation
        sort_by (SortBy): Sort by text or by count

    Returns:
        DefaultDict[str, List[str]]: DefaultDict mapping label to sorted list of the unique
            spans annotated for that label.
    """
    annotations: DefaultDict[str, Set[str]] = defaultdict(set)
    sorted_annotations: DefaultDict[str, List[str]] = defaultdict(list)

    for e in data:
        for s in e.spans:
            span_text = s.text if case_sensitive else s.text.lower()
            annotations[s.label].add(span_text)

    for label, anns in annotations.items():
        sorted_annotations[label] = sorted(anns)

    return sorted_annotations


def get_label_disparities(data: List[Example], label1: str, label2: str, case_sensitive: bool = False) -> Set[str]:
    """Identify annotated spans that have different labels in different examples

    Args:
        data (List[Example]): Input List of examples
        label1 (str): First label to compare
        label2 (str): Second label to compare
        case_sensitive (bool, optional): Consider case of text for each annotation

    Returns:
        Set[str]: Set of all unique text spans that overlap between label1 and label2
    """
    annotations = get_ents_by_label(data, case_sensitive=case_sensitive)
    return set(annotations[label1]).intersection(set(annotations[label2]))


def top_label_disparities(
    data: List[Example], case_sensitive: bool = False, dedupe: bool = False
) -> List[LabelDisparity]:
    """Identify annotated spans that have different labels
    in different examples for all label pairs in data.

    Args:
        data (List[Example]): Input List of examples
        case_sensitive (bool, optional): Consider case of text for each annotation
        dedupe (bool, optional): Whether to deduplicate for table view vs confusion matrix.
            False by default for easy confusion matrix display.

    Returns:
        List[LabelDisparity]: List of LabelDisparity objects for each label pair combination
            sorted by the number of disparities between them.
    """
    annotations = get_ents_by_label(data, case_sensitive=case_sensitive)
    label_disparities = {}
    for label1 in annotations.keys():
        for label2 in annotations.keys():
            if label1 != label2:
                intersection = set(annotations[label1]).intersection(set(annotations[label2]))
                n_disparities = len(intersection)
                if n_disparities > 0:
                    if dedupe:
                        input_hash = "||".join(sorted([label1, label2]))
                    else:
                        input_hash = "||".join([label1, label2])

                    label_disparities[input_hash] = LabelDisparity(label1=label1, label2=label2, count=n_disparities)

    return sorted(label_disparities.values(), key=lambda ld: ld.count, reverse=True)


def top_prediction_errors(
    recognizer: EntityRecognizer,
    data: List[Example],
    labels: List[str] = None,
    n: int = None,
    k: int = None,
    exclude_fp: bool = False,
    exclude_fn: bool = False,
    verbose: bool = False,
) -> List[PredictionError]:
    """Get a sorted list of examples your model is worst at predicting.

    Args:
        recognizer (EntityRecognizer): An instance of EntityRecognizer
        data (List[Example]): List of annotated Examples
        labels (List[str], optional): List of labels to get errors for.
            Defaults to the labels property of `recognizer`.
        n (int, optional): If set, only use the top n examples from data.
        k (int, optional): If set, return the top k prediction errors, otherwise the whole list.
        exclude_fp (bool, optional): Flag to exclude False Positive errors.
        exclude_fn (bool, optional): Flag to exclude False Negative errors.
        verbose (bool, optional): Show verbose output.

    Returns:
        List[PredictionError]: List of Prediction Errors your model is making, sorted by the
            spans your model has the most trouble with.
    """
    labels or recognizer.labels
    if n is not None:
        data = data[:n]

    len(data)
    texts = (e.text for e in data)
    anns = (e.spans for e in data)

    errors = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))  # type: ignore
    error_examples: DefaultDict[Tuple[str, str, str], List[PredictionErrorExamplePair]] = defaultdict(list)
    n_errors = 0

    for orig_example, pred_example, ann in zip(data, recognizer.predict(texts), anns):
        if k is not None and n_errors > k:
            break

        pred_error_example_pair = PredictionErrorExamplePair(original=orig_example, predicted=pred_example)

        cand = set([(s.start, s.end, s.label) for s in pred_example.spans])
        gold = set([(s.start, s.end, s.label) for s in ann])

        fp_diff = cand - gold
        fn_diff = gold - cand

        seen = set()

        if fp_diff and not exclude_fp:
            for fp in fp_diff:
                gold_ent = None
                for ge in gold:
                    if fp[0] == ge[0] and fp[1] == ge[1]:
                        gold_ent = ge
                        break
                if gold_ent:
                    start, end, label = gold_ent
                    text = pred_example.text[start:end]
                    false_label = fp[2]
                    errors[label][text][false_label] += 1
                    error_examples[(text, label, false_label)].append(pred_error_example_pair)
                else:
                    start, end, false_label = fp
                    text = pred_example.text[start:end]
                    errors[NOT_LABELED][text][false_label] += 1
                    error_examples[(text, NOT_LABELED, false_label)].append(pred_error_example_pair)
                n_errors += 1
                seen.add((start, end))

        if fn_diff and not exclude_fn:
            for fn in fn_diff:
                start, end, label = fn
                if (start, end) not in seen:
                    text = pred_example.text[start:end]
                    errors[label][text][NOT_LABELED] += 1
                    error_examples[(text, label, NOT_LABELED)].append(pred_error_example_pair)
                    n_errors += 1

    ranked_errors_map: Dict[Tuple[str, str, str], PredictionError] = {}

    for label, errors_per_label in errors.items():
        for error_text, error_labels in errors_per_label.items():
            for error_label, count in error_labels.items():
                pe_hash = (error_text, label, error_label)
                ranked_errors_map[pe_hash] = PredictionError(
                    text=error_text,
                    true_label=label,
                    pred_label=error_label,
                    count=count,
                    examples=error_examples[pe_hash],
                )

    ranked_errors: List[PredictionError] = sorted(
        list(ranked_errors_map.values()), key=lambda error: error.count, reverse=True  # type: ignore
    )
    error_texts = set()
    for re in ranked_errors:
        if re.examples:
            for e in re.examples:
                error_texts.add(e.original.text)

    error_rate = round(len(error_texts) / len(data), 2)
    if verbose:
        error_summary = {
            "N Examples": len(data),
            "N Errors": len(ranked_errors),
            "N Error Examples": len(error_texts),
            "Error Rate": error_rate,
        }
        msg = Printer()
        msg.divider("Error Analysis")
        msg.table(error_summary)

    return ranked_errors


def get_hardest_examples(
    recognizer: EntityRecognizer,
    examples: List[Example],
    score_count: bool = True,
    normalize_scores: bool = True,
) -> List[HardestExampleV2]:
    """Get hardest examples for a recognizer to predict on and sort by difficulty with the goal
    of quickly identifying the biggest holes in a model / annotated data.

    Args:
        recognizer (EntityRecognizer): EntityRecognizer to test predictions for
        examples (List[Example]): Set of input examples
        score_count (bool): Adjust score by total number of errors
        normalize_scores (bool): Scale scores adjusted by count between 0 and 1

    Returns:
        List[HardestExampleV2]: HardestExamples sorted by difficulty (hardest first)
    """
    preds = recognizer.predict((e.text for e in examples))

    max_count = 0
    hes = []
    for pred, ref in zip(preds, examples):

        scorer = PRFScore()
        scorer.score_set(
            set([(s.start, s.end, s.label) for s in pred.spans]),
            set([(s.start, s.end, s.label) for s in ref.spans]),
        )
        total_errors = scorer.fp + scorer.fn
        score = scorer.fscore if (pred.spans and ref.spans) else 1.0

        if total_errors > max_count:
            max_count = total_errors

        he = HardestExampleV2(reference=ref, prediction=pred, count=total_errors, score=score)
        hes.append(he)

    if score_count:
        for he in hes:
            he.score -= he.count / max_count
        if normalize_scores:
            scores = np.asarray([he.score for he in hes])
            scores = (scores - scores.min()) / np.ptp(scores)
            for i, he in enumerate(hes):
                he.score = scores[i]

    sorted_hes = sorted(hes, key=lambda he: (he.score, he.count))
    return sorted_hes


def get_annotation_labels(examples: List[Example], case_sensitive: bool = False) -> Dict[str, Dict[str, list]]:
    """Constructs a map of each annotation in the list of examples to each label that annotation
    has and references all examples associated with that label.

    Args:
        examples (List[Example]): Input examples
        case_sensitive (bool, optional): Consider case of text for each annotation

    Returns:
        Dict[str, Dict[str, list]]: Annotation map
    """
    annotation_labels_map: Dict[str, Dict[str, list]] = defaultdict(lambda: defaultdict(list))
    for e in examples:
        for s in e.spans:
            text = s.text if case_sensitive else s.text.lower()
            annotation_labels_map[text][s.label].append(e)

    return annotation_labels_map



# isort:skip_file
# type: ignore

from typing import Any, Dict, Iterable, List, Optional, Union

import prodigy
from prodigy.components.db import connect
from prodigy.components.loaders import get_stream
from prodigy.components.preprocess import add_tokens
from prodigy.util import (
    get_labels,
    log,
    split_string,
)
import spacy
from wasabi import msg
from recon.dataset import Dataset
from recon.insights import get_hardest_examples
from recon.operations.core import operation
from recon.recognizer import SpacyEntityRecognizer
from recon.types import Example, HardestExampleV2


@prodigy.recipe(
    "recon.ner_correct",
    # fmt: off
    dataset=("Dataset to save annotations to", "positional", None, str),
    spacy_model=("Loadable spaCy model for tokenization or blank:lang (e.g. blank:en)", "positional", None, str),
    source=("Data to annotate (file path or '-' to read from standard input)", "positional", None, str),
    loader=("Loader (guessed from file extension if not set)", "option", "lo", str),
    label=("Comma-separated label(s) to annotate or text file with one label per line", "option", "l", get_labels),
    exclude=("Comma-separated list of dataset IDs whose annotations to exclude", "option", "e", split_string),
    # fmt: on
)
def ner_correct(
    dataset: str,
    spacy_model: str,
    source: Union[str, Iterable[dict]],
    loader: Optional[str] = None,
    label: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
) -> Dict[str, Any]:

    log("RECIPE: Starting recipe ner.manual", locals())
    nlp = spacy.load(spacy_model)
    labels = label  # comma-separated list or path to text file
    if not labels:
        labels = nlp.pipe_labels.get("ner", [])
        if not labels:
            msg.fail("No --label argument set and no labels found in model", exits=1)
        msg.text(f"Using {len(labels)} labels from model: {', '.join(labels)}")
    log(f"RECIPE: Annotating with {len(labels)} labels", labels)
    stream = get_stream(source, loader=loader, rehash=True, dedup=True, input_key="text")
    # Add "tokens" key to the tasks, either with words or characters
    stream = add_tokens(nlp, stream)
    stream = [Example(**eg) for eg in stream]

    rec = SpacyEntityRecognizer(nlp)
    hes = get_hardest_examples(rec, stream)

    def stream_from_hardest_examples(hes: List[HardestExampleV2]):

        for he in hes:

            combined = he.reference.copy(deep=True)
            pred_spans = []
            for s in he.prediction.spans:
                span = s.copy(deep=True)
                span.label = f"{s.label}:PREDICTED"
                pred_spans.append(span)

            combined.spans = sorted(combined.spans + pred_spans, key=lambda s: s.start)

            task = combined.dict()
            yield task

    stream = stream_from_hardest_examples(hes)

    def before_db(examples):
        for eg in examples:
            new_spans = []
            for span in eg["spans"]:
                if not span["label"].endswith(":PREDICTED"):
                    new_spans.append(span)
            eg["spans"] = new_spans
        return examples

    return {
        "view_id": "blocks",
        "dataset": dataset,
        "stream": stream,
        "exclude": exclude,
        "before_db": before_db,
        "config": {
            "lang": nlp.lang,
            "labels": labels,
            "exclude_by": "input",
            "force_stream_order": True,
            "overlapping_spans": True,
            "blocks": [
                {"view_id": "spans_manual", "overlapping_spans": True},
            ],
            "javascript": f"""

            const disablePredInteraction = event => {{
                let xpath = "//span[contains(text(),':PREDICTED')]";
                let matchingElement = document.evaluate(xpath, document, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
                let length = matchingElement.snapshotLength;

                for (var i = 0; i < length; i++) {{
                    matchingElement.snapshotItem(i).parentElement.parentElement.onclick = function(e) {{ console.log('PRED SPAN CLIECKED'); e.stopImmediatePropagation(); }}
                }}
            }}

            document.addEventListener('prodigyanswer', disablePredInteraction)
            """,
        },
    }


@prodigy.recipe(
    "recon.ner_merge",
    # fmt: off
    dataset=("Dataset with saved annotations to from recon.ner_correct", "positional", None, str),
    recon_dataset=("Recon dataset name", "positional", None, str),
    source=("Source data to merge examples with (file path or List of examples)", "positional", None, str),
    output_dir=("Optional output directory to save dataset to", "positional", None, str),
    exclude=("Comma-separated list of dataset IDs whose annotations to exclude", "option", "e", split_string),
    # fmt: on
)
def ner_merge(
    dataset: str,
    recon_dataset: str,
    source: Union[str, Dataset],
    output_dir: Optional[str] = None,
    exclude: Optional[List[str]] = None,
):
    """
    Stream a List of `recon.types.HardestExample` instances to prodigy
    for review/correction. Uses the Prodigy blocks interface to display
    prediction error information along with ner view
    """
    log("RECIPE: Starting recipe recon.ner_merge", locals())
    if isinstance(source, str):
        dataset = Dataset(recon_dataset).from_disk(source)
    else:
        dataset = source

    DB = connect()
    if dataset not in DB:
        msg.fail(f"Can't find dataset '{dataset}'", exits=1)

    prodigy_raw_examples = DB.get_dataset(dataset)
    prodigy_examples = [Example(**eg) for eg in prodigy_raw_examples if eg["answer"] == "accept"]
    prodigy_texts_to_examples = {e.text: e for e in prodigy_examples}

    prev_len = len(dataset)
    dataset.apply_("recon.v1.prodigy.merge_examples", prodigy_texts_to_examples)
    assert len(dataset) == prev_len

    if output_dir:
        log(f"RECIPE: Fixing {len(prodigy_examples)} examples in data")
        dataset.to_disk(output_dir)


@operation("recon.v1.prodigy.merge_examples")
def merge_examples(example, prodigy_texts_to_examples):
    if example.text in prodigy_texts_to_examples:
        return prodigy_texts_to_examples[example.text]
    else:
        return example



from collections import defaultdict
from typing import Any, Callable, Dict, Iterable, List

import catalogue
import spacy
from recon.linker import BaseEntityLinker, EntityLinker
from recon.types import Entity, Example
from spacy.language import Language


class registry:
    preprocessors = catalogue.create("recon", "preprocessors", entry_points=True)


class preprocessor:
    def __init__(self, name: str):
        """Decorate an operation that makes some changes to a dataset.

        Args:
            name (str): Operation name.
        """
        self.name = name

    def __call__(self, *args: Any, **kwargs: Any) -> Callable:
        """Decorator for an operation.
        The first arg is the function being decorated.
        This function operates on a List[Example].

        e.g. @preprocessor("recon.v1.some_name")

        Or it should operate on a single example and
        recon will take care of applying it to a full Dataset

        Args:
            args: First arg is function to decorate

        Returns:
            Callable: Original function
        """
        op: Callable = args[0]
        registry.preprocessors.register(self.name)(op)

        return op


class PreProcessor:
    def __init__(self, name: str, field: str) -> None:
        self._name = name
        self._field = field
        self._cache: Dict[Any, Any] = {}

    @property
    def name(self) -> str:
        return self._name

    @property
    def field(self) -> str:
        return self._field

    def __call__(self, data: Iterable[Example]) -> Iterable[Any]:
        raise NotImplementedError

    def register(self) -> None:
        registry.preprocessors.register(self.name)(self)


class SpacyPreProcessor(PreProcessor):
    def __init__(self, nlp: Language = None, name: str = "recon.v1.spacy", field: str = "doc") -> None:
        super().__init__(name, field)
        self._nlp = nlp

    @property
    def nlp(self) -> Language:
        if self._nlp is None:
            self._nlp = spacy.blank("en")
            self._nlp.add_pipe("sentencizer")
        return self._nlp

    def __call__(self, data: Iterable[Example]) -> Iterable[Any]:
        unseen_texts = (e.text for i, e in enumerate(data) if hash(e) not in self._cache)
        seen_texts = ((i, e.text) for i, e in enumerate(data) if hash(e) in self._cache)

        docs = list(self.nlp.pipe(unseen_texts))
        for doc in docs:
            self._cache[doc.text] = doc
        for idx, st in seen_texts:
            docs.insert(idx, self._cache[st])

        return docs


class SpanAliasesPreProcessor(PreProcessor):
    def __init__(
        self,
        entities: List[Entity],
        name: str = "recon.v1.span_aliases",
        field: str = "aliases",
        linker: BaseEntityLinker = EntityLinker(),
    ):
        super().__init__(name, field)
        self.entities = entities
        self.ents_to_aliases = defaultdict(list)

        for ent in self.entities:
            if not ent.id:
                ent.id = ent.name
            self.ents_to_aliases[ent.id] = ent.aliases

        self.linker = linker

    def __call__(self, data: Iterable[Example]) -> Iterable[Any]:
        outputs = []

        data = self.linker(data)
        for example in data:
            spans_to_aliases_map = defaultdict(list)
            for span in example.spans:
                if span.kb_id:
                    aliases = self.ents_to_aliases[span.kb_id]
                    spans_to_aliases_map[hash(span)] = aliases

            outputs.append(spans_to_aliases_map)
        return outputs


if "recon.v1.spacy" not in registry.preprocessors:
    registry.preprocessors.register("recon.v1.spacy")(SpacyPreProcessor())



from abc import ABC, abstractmethod
from typing import Iterable, Iterator

from recon.types import Example
from spacy.kb import KnowledgeBase


class BaseEntityLinker(ABC):
    @abstractmethod
    def __call__(self, examples: Iterable[Example]) -> Iterator[Example]:
        raise NotImplementedError


class EntityLinker(BaseEntityLinker):
    def __call__(self, examples: Iterable[Example]) -> Iterator[Example]:
        for example in examples:
            for span in example.spans:
                span.kb_id = span.text
            yield example


class SpacyEntityLinker(BaseEntityLinker):
    def __init__(self, kb: KnowledgeBase):
        self.kb = kb

    def __call__(self, examples: Iterable[Example]) -> Iterator[Example]:
        for example in examples:
            for span in example.spans:
                cands = self.kb.get_candidates(span.text)
                ents = [c.entity_ for c in cands if c.entity_]
                if ents:
                    top_ent = ents[0]
                    span.kb_id = top_ent

            yield example



import functools
import inspect
from collections import OrderedDict
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

from pydantic import BaseConfig, BaseModel
from pydantic.class_validators import Validator
from pydantic.error_wrappers import ErrorWrapper
from pydantic.errors import MissingError
from pydantic.fields import (
    SHAPE_LIST,
    SHAPE_SEQUENCE,
    SHAPE_SET,
    SHAPE_SINGLETON,
    SHAPE_TUPLE,
    SHAPE_TUPLE_ELLIPSIS,
    FieldInfo,
    ModelField,
    Required,
    UndefinedType,
)
from pydantic.schema import get_annotation_from_field_info
from pydantic.typing import ForwardRef, evaluate_forwardref
from pydantic.utils import lenient_issubclass
from recon.types import OperationState

sequence_shapes = {
    SHAPE_LIST,
    SHAPE_SET,
    SHAPE_TUPLE,
    SHAPE_SEQUENCE,
    SHAPE_TUPLE_ELLIPSIS,
}
sequence_types = (list, set, tuple)
sequence_shape_to_type = {
    SHAPE_LIST: list,
    SHAPE_SET: set,
    SHAPE_TUPLE: tuple,
    SHAPE_SEQUENCE: list,
    SHAPE_TUPLE_ELLIPSIS: list,
}


def is_scalar_field(field: ModelField) -> bool:
    field.field_info
    if not (
        field.shape == SHAPE_SINGLETON
        and not lenient_issubclass(field.type_, BaseModel)
        and not lenient_issubclass(field.type_, sequence_types + (dict,))
    ):
        return False
    if field.sub_fields:
        if not all(is_scalar_field(f) for f in field.sub_fields):
            return False
    return True


def is_scalar_sequence_field(field: ModelField) -> bool:
    if (field.shape in sequence_shapes) and not lenient_issubclass(field.type_, BaseModel):
        if field.sub_fields is not None:
            for sub_field in field.sub_fields:
                if not is_scalar_field(sub_field):
                    return False
        return True
    if lenient_issubclass(field.type_, sequence_types):
        return True
    return False


def get_typed_signature(call: Callable[..., Any]) -> inspect.Signature:
    signature = inspect.signature(call)
    globalns = getattr(call, "__globals__", {})
    typed_params = [
        inspect.Parameter(
            name=param.name,
            kind=param.kind,
            default=param.default,
            annotation=get_typed_annotation(param, globalns),
        )
        for param in signature.parameters.values()
    ]
    typed_signature = inspect.Signature(typed_params)
    return typed_signature


def get_typed_annotation(param: inspect.Parameter, globalns: Dict[str, Any]) -> Any:
    annotation = param.annotation
    if isinstance(annotation, str):
        annotation = ForwardRef(annotation)
        annotation = evaluate_forwardref(annotation, globalns, globalns)
    return annotation


def create_response_field(
    name: str,
    type_: Type[Any],
    class_validators: Optional[Dict[str, Validator]] = None,
    default: Optional[Any] = None,
    required: Union[bool, UndefinedType] = False,
    model_config: Type[BaseConfig] = BaseConfig,
    field_info: Optional[FieldInfo] = None,
    alias: Optional[str] = None,
) -> ModelField:
    """
    Create a new response field. Raises if type_ is invalid.
    """
    class_validators = class_validators or {}
    field_info = field_info or FieldInfo(None)

    response_field = functools.partial(
        ModelField,
        name=name,
        type_=type_,
        class_validators=class_validators,
        default=default,
        required=required,
        model_config=model_config,
        alias=alias,
    )

    try:
        return response_field(field_info=field_info)
    except RuntimeError:
        raise ValueError(f"Invalid args for response field! Hint: check that {type_} is a valid pydantic field type")


def get_param_field(
    *,
    param: inspect.Parameter,
    param_name: str,
    default_field_info: Type[FieldInfo] = FieldInfo,
    ignore_default: bool = False,
) -> ModelField:
    default_value = Required
    had_schema = False
    if not param.default == param.empty and ignore_default is False:
        default_value = param.default
    if isinstance(default_value, FieldInfo):
        had_schema = True
        field_info = default_value
        default_value = field_info.default
    else:
        field_info = default_field_info(default_value)
    required = default_value == Required
    annotation: Any = Any
    if not param.annotation == param.empty:
        annotation = param.annotation
    annotation = get_annotation_from_field_info(annotation, field_info, param_name)
    if not field_info.alias and getattr(field_info, "convert_underscores", None):
        alias = param.name.replace("_", "-")
    else:
        alias = field_info.alias or param.name
    field = create_response_field(
        name=param.name,
        type_=annotation,
        default=None if required else default_value,
        alias=alias,
        required=required,
        field_info=field_info,
    )
    field.required = required
    if not had_schema and not is_scalar_field(field=field):
        field.field_info = field_info.default

    return field


def request_body_to_args(
    required_params: List[ModelField],
    received_body: Optional[Dict[str, Any]],
) -> Tuple[Dict[str, Any], List[ErrorWrapper]]:
    values = {}
    errors = []
    if required_params:
        field = required_params[0]
        field_info = field.field_info
        embed = getattr(field_info, "embed", None)
        field_alias_omitted = len(required_params) == 1 and not embed
        if field_alias_omitted:
            received_body = {field.alias: received_body}

        for field in required_params:

            loc: Tuple[str, ...]
            if field_alias_omitted:
                loc = ("body",)
            else:
                loc = ("body", field.alias)

            value: Optional[Any] = None
            if received_body is not None:
                try:
                    value = received_body.get(field.alias)
                except AttributeError:
                    errors.append(get_missing_field_error(loc))
                    continue
            if value is None:
                if field.required:
                    errors.append(get_missing_field_error(loc))
                else:
                    values[field.name] = deepcopy(field.default)
                continue

            v_, errors_ = field.validate(value, values, loc=loc)

            if isinstance(errors_, ErrorWrapper):
                errors.append(errors_)
            elif isinstance(errors_, list):
                errors.extend(errors_)
            else:
                values[field.name] = v_
    return values, errors


def get_missing_field_error(loc: Tuple[str, ...]) -> ErrorWrapper:
    missing_field_error = ErrorWrapper(MissingError(), loc=loc)
    return missing_field_error


def get_required_operation_params(op: Callable) -> Dict[str, ModelField]:
    """Get required typed parameters for an operation

    Based on logic for JSON resolution to Pydantic types implemented by FastAPI
    Reference: https://github.com/tiangolo/fastapi/blob/master/fastapi/dependencies/utils.py

    Args:
        op (Callable): Inner Callable of a recon operation

    Returns:
        Dict[str, ModelField]: Mapping of field name to Pydantic ModelField for
            required parameters of the callable
    """

    endpoint_signature = get_typed_signature(op)
    signature_params = endpoint_signature.parameters

    required_params = OrderedDict()
    for param_name, param in signature_params.items():
        if param_name in {"example", "preprocessed_outputs"}:
            # All operations accept an example and potentially a special
            # preprocessed_outputs parameter if the operation requires
            # batched example preprocessing to function. We don't
            # need to include these in dynamic param resolution as they are
            # already handled by the internals of the operation
            continue

        param_field = get_param_field(param=param, default_field_info=FieldInfo, param_name=param_name)
        required_params[param_name] = param_field

    return required_params


def get_received_operation_data(required_params: Dict[str, ModelField], state: OperationState) -> Dict[str, Any]:
    """Resolve serialized args and kwargs data of an operation to their Pydantic types

    Args:
        required_params (Dict[str, ModelField]): Mapping of field name to Pydantic ModelField for
            required parameters of the callable
        state (OperationState): Operation state including serialized args and kwargs data

    Returns:
        Dict[str, Any]: Kwargs mapping of deserialized params to pass to the operation
    """

    filled_keys = set()
    received_data = {}

    if state.kwargs:
        for key, val in state.kwargs.items():
            if key in required_params:
                received_data[key] = val
                filled_keys.add(key)

    if set(required_params.keys()) - filled_keys and state.args:
        for arg_idx, arg in enumerate(state.args):
            param_field = required_params[list(required_params.keys())[arg_idx]]
            received_data[param_field.name] = arg

    return received_data



from pathlib import Path
from typing import cast

from recon.corpus import Corpus
from recon.stats import calculate_label_distribution_similarity, get_ner_stats
from recon.types import NERStats
from wasabi import Printer


def stats(data_dir: Path) -> None:
    """Calculate statistics on a Corpus

    Args:
        data_dir (Path): Path to data folder
    """

    msg: Printer = Printer()

    def print_stats(corpus: Corpus) -> None:
        for ds, ner_stats in corpus.apply(get_ner_stats).items():
            ner_stats = cast(NERStats, ner_stats)
            sorted_labels = sorted(ner_stats.n_annotations_per_type.keys())

            msg.text(f"Stats for {ds.capitalize()} data")
            msg.text("--------------------")
            msg.table(
                {
                    "N Examples": ner_stats.n_examples,
                    "N Annotations": ner_stats.n_annotations,
                }
            )
            msg.info(f"Labels in {ds}")
            msg.text(sorted_labels)

            msg.info("N Annotations per Label")
            msg.table(ner_stats.n_annotations_per_type)

    with msg.loading("Loading Corpus from Disk"):
        corpus = Corpus.from_disk(data_dir)
        msg.good("Done")

    msg.divider("Calculating stats")

    print_stats(corpus)
    train_dev_sim = calculate_label_distribution_similarity(corpus.train, corpus.dev)
    train_test_sim = calculate_label_distribution_similarity(corpus.train, corpus.test)
    dev_test_sim = calculate_label_distribution_similarity(corpus.dev, corpus.test)

    msg.divider("Calculating Similarity between label distributions of Corpus")

    msg.table(
        {
            "Train / Dev": round(train_dev_sim, 2),
            "Train / Test": round(train_test_sim, 2),
            "Dev   / Test": round(dev_test_sim, 2),
        }
    )



import random
from collections import defaultdict
from typing import Any, Dict, List, Tuple

from recon.types import Example


def hash_example_meta(example: Example, fields: List[str] = None, ignore_field_absence: bool = False) -> Tuple:
    """Create a hash out of the metadata of an example

    Args:
        example (Example): Input Example
        fields (List[str], optional): Meta fields to use in hash. Defaults to all availabile fields.
        ignore_field_absence (bool, optional): Determines behavior of when a field does not exist
            in an example's meta property

    Raises:
        ValueError: If a field passed to fields does not exist in an example's meta

    Returns:
        Tuple: Tuple of hashable attributes from the meta of an example
    """
    if fields is None:
        fields = list(example.meta.keys())

    tpl = []
    for field in fields:
        tpl.append(field)
        if field not in example.meta:
            if ignore_field_absence:
                continue
            raise ValueError(f"Field {field} not present in 'meta' for example {example}")
        meta_val = example.meta[field]
        if isinstance(meta_val, list):
            tpl += meta_val
        else:
            tpl.append(meta_val)

    return tuple(tpl)


def sample_examples(
    examples: List[Example],
    meta_filters: Dict[str, List[str]] = {},
    fields: List[str] = None,
    ignore_field_absence: bool = False,
    top_k_per_hash: int = 10,
    top_k: int = -1,
    shuffle: bool = True,
) -> List[Example]:
    """Sample examples based on meta attributes

    Args:
        examples (List[Example]): Examples to sample from
        meta_filters (Dict[str, List[str]], optional): Values to filter out of sampled set for each meta field.
        fields (List[str], optional): Meta fields to use in hash. Defaults to all availabile fields.
        ignore_field_absence (bool, optional): Determines behavior of when a field does not exist
            in an example's meta property.
        top_k_per_hash (int, optional): Number of examples to include per meta hash.
        top_k (int, optional): Total number of examples to sample from.

    Returns:
        List[Example]: Sampled examples
    """

    examples_counter: Dict[Any, int] = defaultdict(int)
    out_examples = []

    if shuffle:
        random.shuffle(examples)

    for example in examples:

        if top_k > 0 and sum(examples_counter.values()) >= top_k:
            break

        for meta_field, meta_vals in meta_filters.items():
            if example.meta[meta_field] not in meta_vals:
                break

        meta_hash = hash_example_meta(example, fields=fields, ignore_field_absence=ignore_field_absence)
        if examples_counter[meta_hash] <= top_k_per_hash:
            out_examples.append(example)
            examples_counter[meta_hash] += 1

    return out_examples



import tempfile
from pathlib import Path
from typing import Iterable, Iterator, List, Set

from recon.loaders import to_spacy
from recon.types import Example, Scores, Span, Token
from spacy.language import Language
from spacy.training import Corpus as SpacyCorpus
from wasabi import Printer


class EntityRecognizer:
    """Abstract Base class for recognizing entities in a batch of text.
    Used in the `recon.insights` module for understanding the kinds
    of examples your model is having the most trouble with.
    """

    @property
    def labels(self) -> List[str]:
        """Return List of String Labels

        Raises:
            NotImplementedError: Not Implemented, override

        Returns:
            List[str]: List of labels the model can predict
        """
        raise NotImplementedError

    def predict(self, texts: Iterable[str]) -> Iterator[Example]:
        """Run model inference on a batch of raw texts.

        Args:
            texts (Iterable[str]): Raw text examples

        Raises:
            NotImplementedError: Not implemented, override

        Returns:
            Iterator[Example]: Iterator of Examples
        """
        raise NotImplementedError

    def _evaluate(self, examples: List[Example]) -> Scores:
        raise NotImplementedError

    def evaluate(self, examples: List[Example], verbose: bool = True) -> Scores:
        """Evaluate recognizer performance on dataset and print metrics

        Args:
            data (List[Example]): Examples to evaluate on
            verbose (bool, optional): Print results or not. Defaults to True.

        Returns:
            Scorer: spaCy scorer object
        """
        sc = self._evaluate(examples)

        msg = Printer(no_print=not verbose)
        msg.divider("Recognizer Results")
        result = [
            ("Precision", f"{sc.ents_p:.3f}"),
            ("Recall", f"{sc.ents_r:.3f}"),
            ("F-Score", f"{sc.ents_f:.3f}"),
        ]
        msg.table(result)

        table_data = []
        for label, scores in sorted(sc.ents_per_type.items(), key=lambda tup: tup[0]):
            table_data.append((label, f"{scores['p']:.3f}", f"{scores['r']:.3f}", f"{scores['f']:.3f}"))
        header = ("Label", "Precision", "Recall", "F-Score")
        formatted = msg.table(table_data, header=header, divider=True)
        return sc


class SpacyEntityRecognizer(EntityRecognizer):
    """Create an EntityRecognizer from a spaCy Langauge instance"""

    def __init__(self, nlp: Language):
        """Initialize a SpacyEntityRecognizer

        Args:
            nlp (Language): spaCy Language instance that can sets doc.ents
        """
        super().__init__()
        self.nlp = nlp

    @property
    def labels(self) -> List[str]:
        """Return List of spaCy ner labels

        Returns:
            List[str]: List of labels from spaCy ner pipe
        """
        all_labels: Set[str] = set()

        for pipe in ["ner", "entity_ruler"]:
            if self.nlp.has_pipe(pipe):
                all_labels = all_labels | set(self.nlp.get_pipe(pipe).labels)

        return sorted(list(all_labels))

    def predict(self, texts: Iterable[str]) -> Iterator[Example]:
        """Run spaCy nlp.pipe on a batch of raw texts.

        Args:
            texts (Iterable[str]): Raw text examples

        Yields:
            Iterator[Example]: Examples constructed from spaCy Model predictions
        """
        examples: List[Example] = []

        for doc in self.nlp.pipe(texts):
            yield Example(
                text=doc.text,
                spans=[
                    Span(
                        text=e.text,
                        start=e.start_char,
                        end=e.end_char,
                        label=e.label_,
                        token_start=e.start,
                        token_end=e.end,
                    )
                    for e in doc.ents
                ],
                tokens=[Token(text=t.text, start=t.idx, end=t.idx + len(t), id=t.i) for t in doc],
            )

    def _evaluate(self, data: List[Example]) -> Scores:
        """Evaluate spaCy recognizer performance on dataset

        Args:
            data (List[Example]): Examples to evaluate on
            verbose (bool, optional): Print results or not. Defaults to True.

        Returns:
            Scorer: spaCy scorer object
        """

        with tempfile.TemporaryDirectory() as tmp_dir:
            data_path = Path(tmp_dir) / "data.spacy"
            to_spacy(data_path, data)
            corpus = SpacyCorpus(data_path, gold_preproc=False)
            dev_dataset = list(corpus(self.nlp))
            sc = self.nlp.evaluate(dev_dataset)
            scores = Scores(**sc)
        return scores



from typing import Any, Callable, Tuple, Union

import xxhash


def token_hash(token: Any, as_int: bool = True) -> Union[str, int]:
    """Hash of Token type

    Args:
        token (Token): Token to hash
        as_int (bool, optional): Encode hash as int

    Returns:
        Union[str, int]: Token hash
    """
    return _hash((token.text, token.start, token.end, token.id), as_int=as_int)


def span_hash(span: Any, as_int: bool = True) -> Union[str, int]:
    """Hash of Span type

    Args:
        span (Span): Span to hash
        as_int (bool, optional): Encode hash as int

    Returns:
        Union[str, int]: Span hash
    """
    hash_data = (
        span.start,
        span.end,
        span.label,
        span.text,
        span.token_start if span.token_start else 0,
        span.token_end if span.token_end else 0,
    )
    return _hash(hash_data, as_int=as_int)


def example_hash(example: Any, as_int: bool = True) -> Union[str, int]:
    """Hash of Example type

    Args:
        example (Example): Example to hash
        as_int (bool, optional): Encode hash as int

    Returns:
        Union[str, int]: Example hash
    """
    hash_data = (example.text,) + tuple((span_hash(span, as_int=False) for span in example.spans))
    return _hash(hash_data, as_int=as_int)


def tokenized_example_hash(example: Any, as_int: bool = True) -> Union[str, int]:
    """Hash of Example type including token data

    Args:
        example (Example): Example to hash
        as_int (bool, optional): Encode hash as int

    Returns:
        Union[str, int]: Example hash
    """
    tokens = example.tokens or []
    hash_data = (
        (example.text,)
        + tuple((span_hash(span, as_int=False) for span in example.spans))
        + tuple((token_hash(token, as_int=False) for token in tokens))
    )
    return _hash(hash_data, as_int=as_int)


def dataset_hash(dataset: Any, as_int: bool = True) -> Union[str, int]:
    """Hash of Dataset

    Args:
        dataset (Dataset): Dataset to hash
        as_int (bool, optional): Encode hash as int

    Returns:
        Union[str, int]: Dataset hash
    """
    hash_data = (dataset.name,) + tuple((example_hash(example, as_int=False) for example in dataset.data))
    return _hash(hash_data, as_int=as_int)


def prediction_error_hash(prediction_error: Any, as_int: bool = True) -> Union[str, int]:
    """Hash of PredictionError

    Args:
        prediction_error (PredictionError): PredictionError to hash
        as_int (bool, optional): Encode hash as int

    Returns:
        Union[str, int]: PredictionError hash
    """
    hash_data = (prediction_error.text, prediction_error.true_label, prediction_error.pred_label)
    return _hash(hash_data, as_int=as_int)


def _hash(tpl: Tuple, hash_function: Callable = xxhash.xxh64, as_int: bool = True) -> Union[str, int]:
    """Deterministic hash function. The main use here is
    providing a `commit_hash` for a Dataset to compare across
    saves/loads and ensure that operations are re-run if the hash
    ever changes

    Args:
        tpl (Tuple): Tuple of data to hash
        hash_function (Callable, hashlib.sha1): Hash function from
            python hashlib. Defaults to sha1 (same as git)
        as_int (bool, optional): Encode hash as int

    Returns:
        Union[str, int]: Deterministic hash using tpl data
    """
    m = hash_function()
    for e in tpl:
        if isinstance(e, str):
            e_bytes = e.encode("utf-8")
        else:
            e_bytes = bytes(e)
        m.update(e_bytes)
    return m.intdigest() if as_int else m.hexdigest()



"""ReconNER, Debug annotated Named Entity Recognition (NER) data for inconsitencies and get insights on improving the quality of your data."""

__version__ = "0.9.1"

from recon.augmentation import *
from recon.corpus import Corpus
from recon.corrections import *
from recon.insights import *
from recon.loaders import read_json, read_jsonl
from recon.operations.core import operation
from recon.preprocess import SpacyPreProcessor
from recon.stats import get_ner_stats
from recon.tokenization import add_tokens, fix_tokenization_and_spacing
from recon.validation import filter_overlaps, upcase_labels

try:
    # This needs to be imported in order for the entry points to be loaded
    from recon.prodigy import recipes as prodigy_recipes  # noqa: F401
except ImportError as e:
    print("Error", e)
    pass



from typing import List

from recon.types import Example


def to_prodigy(
    examples: List[Example],
    prodigy_dataset: str,
    overwrite_dataset: bool = False,
    add_hash: bool = True,
) -> None:
    """Save a list of examples to Prodigy

    Args:
        examples (List[Example]): Input examples
        prodigy_dataset (str): Name of Prodigy dataset to save to
        overwrite_dataset (bool, optional): If True will overwrite all data in Prodigy

    Raises:
        ValueError: If trying to save examples to an existing dataset without explicitly
            setting overwrite_dataset to True
    """
    from prodigy.core import connect
    from prodigy.util import set_hashes

    db = connect()

    if db.get_dataset(prodigy_dataset):
        if overwrite_dataset:
            db.drop_dataset(prodigy_dataset)
            db.add_dataset(prodigy_dataset)
        else:
            raise ValueError(f"Prodigy dataset {prodigy_dataset} already exists.")
    else:
        db.add_dataset(prodigy_dataset)

    prodigy_examples = []
    for e in examples:
        prodigy_examples.append(set_hashes(e.dict()))

    db.add_examples(prodigy_examples, [prodigy_dataset])


def from_prodigy(prodigy_dataset: str) -> List[Example]:
    """Load Recon examples from a Prodigy dataset.

    Args:
        prodigy_dataset (str): Name of prodigy dataset to load from

    Raises:
        ValueError: If trying to load examples from a dataset that doesn't exist in prodigy

    Returns:
        List[Example]: List of Recon examples
    """

    from prodigy.core import connect

    db = connect()

    examples = []
    prodigy_examples = db.get_dataset(prodigy_dataset)
    if not prodigy_examples:
        raise ValueError(
            f"Prodigy dataset with name {prodigy_dataset} does not exist. Available datasets are: \n {', '.join(db.datasets)}"
        )
    examples = [Example(**e) for e in prodigy_examples]
    return examples



from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, List, Set, Union, cast

import srsly
from recon.hashing import dataset_hash
from recon.loaders import from_spacy, read_json, read_jsonl, to_spacy
from recon.operations import registry
from recon.stats import get_ner_stats
from recon.store import ExampleStore
from recon.types import (
    DatasetOperationsState,
    Example,
    OperationResult,
    OperationState,
    OperationStatus,
    TransformationType,
)
from spacy.util import ensure_path
from wasabi import Printer


class Dataset:
    """A Dataset is a around a List of examples.
    Datasets are responsible for tracking all Operations done on them.
    This ensures data lineage and easy reporting of how changes in the
    data based on various Operations effects overall quality.

    Dataset holds state (let's call it self.operations for now)
    self.operations is a list of every function run on the Dataset since it's
    initial creation. If loading from disk, track everything that happens in loading
    phase in operations as well by simply initializing self.operations in constructors

    Each operation should has the following attributes:
        operation hash
        name: function/callable name ideally, could be added with a decorator
        status: (not_started|completed)
        transformations: List[Transformation]
            commit hash
            timestamp(s) - start and end both? end is probably enough
            examples deleted
            examples added
            examples corrected
            annotations deleted
            annotations added
            annotations corrected

            for annotations deleted/added/corrected, include mapping from old Example hash to new Example hash
            that can be decoded for display later

    All operations are serializable in the to_disk and from_disk methods.

    So if I have 10 possible transformations.

    I can run 1..5, save to disk train a model and check results.
    Then I can load that model from disk with all previous operations already tracked
    in self.operations. Then I can run 6..10, save to disk and train model.
    Now I have git-like "commits" for the data used in each model.
    """

    def __init__(
        self,
        name: str,
        data: List[Example] = [],
        operations: List[OperationState] = None,
        example_store: ExampleStore = None,
        version: str = "0.0.0",
        verbose: bool = True,
    ):
        self._name = name
        self.data = data
        if not operations:
            operations = []
        self.operations = operations

        if example_store is None:
            example_store = ExampleStore(data)
        self.example_store = example_store
        self._version = version
        self.verbose = verbose

    @property
    def name(self) -> str:
        return self._name

    @property
    def version(self) -> str:
        return self._version

    @property
    def commit_hash(self) -> str:
        return cast(str, dataset_hash(self, as_int=False))

    def __str__(self) -> str:
        stats = get_ner_stats(self.data, serialize=True)
        return f"Dataset: {self.name}\n" f"Stats: {stats}"

    def __hash__(self) -> int:
        return cast(int, dataset_hash(self))

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, example_hash: int) -> Example:
        for e in self.data:
            if hash(e) == example_hash:
                return e
        raise KeyError(f"Example with hash {example_hash} does not exist")

    def apply(self, func: Callable[[List[Example], Any, Any], Any], *args: Any, **kwargs: Any) -> Any:
        """Apply a function to the dataset

        Args:
            func (Callable[[List[Example], Any, Any], Any]):
                Function from an existing recon module that can operate on a List of examples

        Returns:
            Result of running func on List of examples
        """
        return func(self.data, *args, **kwargs)  # type: ignore

    def apply_(
        self,
        operation: Union[str, Callable[[Any], OperationResult]],
        *args: Any,
        initial_state: OperationState = None,
        **kwargs: Any,
    ) -> None:
        """Apply an operation to all data inplace.

        Args:
            operation (Callable[[Any], OperationResult]): Any operation that
                changes data in place. See recon.operations.registry.operations
        """
        if isinstance(operation, str):
            operation = registry.operations.get(operation)
            if operation:
                operation = cast(Callable, operation)

        name = getattr(operation, "name", None)
        if not name:
            raise ValueError("This function is not an operation since it does not have a name.")

        msg = Printer(no_print=self.verbose == False)
        msg.text(f"=> Applying operation '{name}' to dataset '{self.name}'")
        result: OperationResult = operation(self, *args, initial_state=initial_state, verbose=self.verbose, **kwargs)  # type: ignore
        msg.good(f"Completed operation '{name}'")

        self.operations.append(result.state)
        dataset_changed = any(
            (
                result.state.examples_added,
                result.state.examples_removed,
                result.state.examples_changed,
            )
        )
        if dataset_changed:
            self.data = result.data

    def pipe_(self, operations: List[Union[str, OperationState]]) -> None:
        """Run a sequence of operations on dataset data.
        Internally calls Dataset.apply_ and will resolve named
        operations in registry.operations

        Args:
            operations (List[Union[str, OperationState]]): List of operations
        """

        msg = Printer(no_print=self.verbose == False)
        msg.text(f"Applying pipeline of operations inplace to the dataset: {self.name}")

        for op in operations:
            op_name = op.name if isinstance(op, OperationState) else op
            msg.text(f"|_ {op_name}")

        for op in operations:
            if isinstance(op, str):
                op_name = op
                args = []
                kwargs = {}
                initial_state = None
            elif isinstance(op, OperationState):
                op_name = op.name
                args = op.args
                kwargs = op.kwargs
                initial_state = op

            operation = registry.operations.get(op_name)

            self.apply_(operation, *args, initial_state=initial_state, **kwargs)

    def rollback(self, n: int = 1) -> None:
        """Rollback the last n operations on a dataset.

        e.g.
            ```
            ds = Dataset("name", data)

            initial_ds_hash = hash(ds)

            ds.apply_("some_operation")
            ds.rollback()

            hash(ds) == initial_ds_hash
            >>> True # This should be True

        Args:
            n (int): Number of operations to rollback
        """

        if n < 1:
            raise ValueError("Cannot rollback dataset: n must be 1 or higher.")
        elif n > len(self.operations):
            raise ValueError("Cannot rollback dataset: n is larger than the total number of dataset operations.")

        store = self.example_store

        examples_to_remove = set()
        examples_to_add = []

        for op in self.operations[-n:]:
            for t in op.transformations:
                if t.type == TransformationType.EXAMPLE_ADDED:
                    examples_to_remove.add(t.example)
                elif t.type == TransformationType.EXAMPLE_CHANGED:
                    examples_to_remove.add(t.example)
                    examples_to_add.append(store[t.prev_example])  # type: ignore
                elif t.type == TransformationType.EXAMPLE_REMOVED:
                    examples_to_add.append(store[t.prev_example])  # type: ignore

        old_data = [e for e in self.data if hash(e) not in examples_to_remove]
        old_data += examples_to_add

        self.data = old_data
        self.operations = self.operations[:-1]
        for e in examples_to_remove:
            del self.example_store._map[e]  # type: ignore

    def search(self, search_query: str, case_sensitive: bool = True) -> List[Example]:
        """Naive search method to quickly identify examples matching the provided substring

        Args:
            search_query (str): Substring to search each example for
            case_sensitive (bool, optional): Consider case of search query and example text

        Returns:
            List[Example]: Matched examples
        """
        search_query = search_query if case_sensitive else search_query.lower()
        out_examples = []

        for example in self.data:
            example_text = example.text if case_sensitive else example.text.lower()
            if search_query in example_text:
                out_examples.append(example)

        return out_examples

    def from_disk(self, path: Path, loader_func: Callable = read_jsonl) -> "Dataset":
        """Load Dataset from disk given a path and a loader function that reads the data
        and returns an iterator of Examples

        Args:
            path (Path): path to load from
            loader_func (Callable, optional): Callable that reads a file and returns a List of examples.
                Defaults to [read_jsonl][recon.loaders.read_jsonl]
        """
        path = ensure_path(path)
        state = None
        if (path / ".recon" / self.name).exists():
            state = srsly.read_json(path / ".recon" / self.name / "state.json")
            state = DatasetOperationsState(**state)
            self.operations = state.operations

            example_store_path = path / ".recon" / self.name / "example_store.jsonl"
            if example_store_path.exists():
                self.example_store.from_disk(example_store_path)

        if loader_func == read_json:
            suffix = ".json"
        else:
            suffix = ".jsonl"

        data = loader_func(path / f"{self.name}{suffix}")
        self.data = data

        for example in self.data:
            self.example_store.add(example)

        if state and self.commit_hash != state.commit:
            # Dataset changed, examples added
            self.operations.append(
                OperationState(
                    name="recon.v1.examples_added_external",
                    status=OperationStatus.COMPLETED,
                    ts=datetime.now(),
                    examples_added=max(len(self) - state.size, 0),
                    examples_removed=max(state.size - len(self), 0),
                    examples_changed=0,
                    transformations=[],
                )
            )

            for op in self.operations:
                op.status = OperationStatus.NOT_STARTED

        seen: Set[str] = set()
        operations_to_run: Dict[str, OperationState] = {}

        for op in self.operations:
            if (
                op.name not in operations_to_run
                and op.name in registry.operations
                and op.status != OperationStatus.COMPLETED
            ):
                operations_to_run[op.name] = op

        for op_name, state in operations_to_run.items():
            op = registry.operations.get(op_name)

            self.apply_(op, *state.args, initial_state=state, **state.kwargs)  # type: ignore

        return self

    def to_disk(self, output_dir: Path, force: bool = False, save_examples: bool = True) -> None:
        """Save Corpus to Disk

        Args:
            output_dir (Path): Output file path to save data to
            force (bool): Force save to directory. Create parent directories
                or overwrite existing data.
            save_examples (bool): Save the example store along with the state.
        """
        output_dir = ensure_path(output_dir)
        state_dir = output_dir / ".recon" / self.name
        if force:
            output_dir.mkdir(parents=True, exist_ok=True)

            if not state_dir.exists():
                state_dir.mkdir(parents=True, exist_ok=True)

        state = DatasetOperationsState(
            name=self.name, commit=self.commit_hash, size=len(self), operations=self.operations
        )
        srsly.write_json(state_dir / "state.json", state.dict())

        if save_examples:
            self.example_store.to_disk(state_dir / "example_store.jsonl")

        srsly.write_jsonl(output_dir / (self.name + ".jsonl"), [e.dict() for e in self.data])

    def from_prodigy(self, prodigy_datasets: List[str]) -> "Dataset":
        """Need to have from_prodigy accept multiple datasets as a list of str so Prodigy
        can stay separate and new annotation sessions can happen often. Basically prodigy db-merge

        Need to save to only 1 prodigy dataset though for consistency

        Args:
            prodigy_datasets (List[str]): List of prodigy datasets to load from

        Returns:
            Dataset: Initialized dataset with Prodigy data
        """
        from recon.prodigy.utils import from_prodigy

        print(f"Loading data from prodigy datasets: {', '.join(prodigy_datasets)}")
        data = []
        for prodigy_dataset in prodigy_datasets:
            data += from_prodigy(prodigy_dataset)
        self.data = data
        return self

    def to_prodigy(self, prodigy_dataset: str = None, overwrite: bool = True) -> str:
        """Save examples to prodigy dataset

        Args:
            prodigy_dataset (str, optional): Prodigy dataset name to save to.
            overwrite (bool, optional): Overwrite dataset name if it exists.

        Returns:
            str: Prodigy dataset name
        """
        from recon.prodigy.utils import to_prodigy

        if not prodigy_dataset:
            prodigy_dataset = f"{self.name}_{self.commit_hash}"

        print(f"Saving dataset to prodigy dataset: {prodigy_dataset}")
        to_prodigy(self.data, prodigy_dataset, overwrite_dataset=overwrite)
        return prodigy_dataset

    def from_spacy(self, path: Path) -> "Dataset":
        """Load Dataset from a file in the .spacy format

        Args:
            path (Path): path to load from

        Returns:
            Dataset: Initialized dataset with Prodigy data
        """
        data = from_spacy(path)
        self.data = list(data)
        return self

    def to_spacy(self, output_dir: Path) -> None:
        """Save data to .spacy file

        Saves file as {output_dir}/{self.name}.spacy

        Args:
            output_dir (Path): Output file path to save data to
        """
        output_dir = ensure_path(output_dir)
        to_spacy(output_dir / (self.name + ".spacy"), self.data)



from pathlib import Path
from typing import Any, Callable, List, Tuple, Union

import srsly
from recon.dataset import Dataset
from recon.loaders import read_jsonl
from recon.store import ExampleStore
from recon.types import (
    CorpusApplyResult,
    CorpusMeta,
    Example,
    OperationResult,
    OperationState,
)
from spacy.util import ensure_path


class Corpus:
    """Container for a full Corpus with train/dev/test splits.
    Used to apply core functions to all datasets at once.
    """

    def __init__(
        self,
        name: str,
        train: Dataset,
        dev: Dataset,
        test: Dataset = None,
        example_store: ExampleStore = None,
    ):
        """Initialize a Corpus.

        Args:
            name (str): Name of the Corpus
            train (Dataset): List of examples for **train** set
            dev (Dataset): List of examples for **dev** set
            test (Dataset, optional): Defaults to None. List of examples for **test** set
        """
        self._name = name
        if example_store is None:
            examples = train.data + dev.data
            if test:
                examples += test.data
            example_store = ExampleStore(examples)
        self.example_store = example_store

        if test is None:
            test = Dataset("test")

        for ds in (train, dev, test):
            ds.example_store = example_store

        self._train = train
        self._dev = dev
        self._test = test

    @property
    def name(self) -> str:
        """Get Corpus name

        Returns:
            str: Corpus name
        """
        return self._name

    @property
    def train_ds(self) -> Dataset:
        """Return train dataset

        Returns:
            Dataset: Train Dataset
        """
        return self._train

    @property
    def dev_ds(self) -> Dataset:
        """Return dev dataset

        Returns:
            Dataset: Dev Dataset
        """
        return self._dev

    @property
    def test_ds(self) -> Dataset:
        """Return test dataset

        Returns:
            Dataset: Test Dataset
        """
        return self._test

    @property
    def train(self) -> List[Example]:
        """Return train dataset

        Returns:
            List[Example]: Train Examples
        """
        return self._train.data

    @property
    def dev(self) -> List[Example]:
        """Return dev dataset

        Returns:
            List[Example]: Dev Examples
        """
        return self._dev.data

    @property
    def test(self) -> List[Example]:
        """Return test dataset

        Returns:
            List[Example]: Test Examples
        """
        return self._test.data or []

    @property
    def all(self) -> List[Example]:
        """Return concatenation of train/dev/test datasets

        Returns:
            List[Example]: All Examples in Corpus
        """
        return self.train + self.dev + self.test

    def apply(self, func: Callable[[List[Example], Any, Any], Any], *args: Any, **kwargs: Any) -> CorpusApplyResult:
        """Apply a function to all datasets

        Args:
            func (Callable[[List[Example], Any, Any], Any]):
                Function from an existing recon module that can operate on a List of examples

        Returns:
            CorpusApplyResult: CorpusApplyResult mapping dataset name to return type of func Callable
        """

        return CorpusApplyResult(
            train=func(self.train, *args, **kwargs),  # type: ignore
            dev=func(self.dev, *args, **kwargs),  # type: ignore
            test=func(self.test, *args, **kwargs),  # type: ignore
            all=func(self.all, *args, **kwargs),  # type: ignore
        )

    def apply_(self, operation: Callable[[Any], OperationResult], *args: Any, **kwargs: Any) -> None:
        """Apply a function to all data inplace.

        Args:
            operation (Callable[[Any], OperationResult]): Any operation that
                changes data in place. See recon.operations.registry.operations
        """
        self._train.apply_(operation, *args, **kwargs)
        self._dev.apply_(operation, *args, **kwargs)
        self._test.apply_(operation, *args, **kwargs)

    def pipe_(self, operations: List[Union[str, OperationState]]) -> None:
        """Run a sequence of operations on each dataset.
        Calls Dataset.pipe_ for each dataset

        Args:
            operations (List[Union[str, OperationState]]): List of operations
        """
        self._train.pipe_(operations)
        self._dev.pipe_(operations)
        self._test.pipe_(operations)

    @classmethod
    def from_disk(
        cls,
        data_dir: Path,
        name: str = "corpus",
        train_name: str = "train",
        dev_name: str = "dev",
        test_name: str = "test",
        loader_func: Callable = read_jsonl,
    ) -> "Corpus":
        """Load Corpus from disk given a directory with files
        named explicitly train.jsonl, dev.jsonl, and test.jsonl

        Args:
            data_dir (Path): directory to load from.
            train_name (str, optional): Name of train data under data_dir. Defaults to train.
            dev_name (str, optional): Name of dev data under data_dir. Defaults to dev.
            test_name (str, optional): Name of test data under data_dir. Defaults to test.
            loader_func (Callable, optional): Callable that reads a file and returns a List of examples.
                Defaults to [read_jsonl][recon.loaders.read_jsonl]
        """
        data_dir = ensure_path(data_dir)

        corpus_meta_path = data_dir / ".recon" / "meta.json"
        if corpus_meta_path.exists():
            corpus_meta = CorpusMeta.parse_file(corpus_meta_path)
            name = corpus_meta.name

        example_store_path = data_dir / ".recon" / "example_store.jsonl"
        example_store = ExampleStore()
        if example_store_path.exists():
            example_store.from_disk(example_store_path)

        train = Dataset("train", example_store=example_store).from_disk(data_dir)
        dev = Dataset("dev", example_store=example_store).from_disk(data_dir)

        try:
            test = Dataset("test", example_store=example_store).from_disk(data_dir)
            corpus = cls(name, train, dev, test=test)
        except ValueError:
            corpus = cls(name, train, dev)
        return corpus

    def to_disk(self, data_dir: Path, force: bool = False) -> None:
        """Save Corpus to Disk

        Args:
            data_dir (Path): Directory to save data to
            force (bool): Force save to directory. Create parent directories
                or overwrite existing data.
        """
        data_dir = ensure_path(data_dir)
        state_dir = data_dir / ".recon"
        corpus_meta_path = state_dir / "meta.json"

        if force:
            data_dir.mkdir(parents=True, exist_ok=True)

            if not state_dir.exists():
                state_dir.mkdir(parents=True, exist_ok=True)

        srsly.write_json(corpus_meta_path, CorpusMeta(name=self.name).dict())
        self._train.to_disk(data_dir, force=force, save_examples=False)
        self._dev.to_disk(data_dir, force=force, save_examples=False)
        if self._test:
            self._test.to_disk(data_dir, force=force, save_examples=False)

        self.example_store.to_disk(state_dir / "example_store.jsonl")

    @classmethod
    def from_prodigy(
        cls,
        name: str,
        prodigy_train_datasets: List[str],
        prodigy_dev_datasets: List[str],
        prodigy_test_datasets: List[str] = None,
    ) -> "Corpus":
        """Load a Corpus from 3 separate datasets in Prodigy

        Args:
            prodigy_train_dataset (str, optional): Train dataset name in Prodigy
            prodigy_dev_dataset (str, optional): Dev dataset name in Prodigy
            prodigy_test_dataset (str, optional): Test dataset name in Prodigy

        Returns:
            Corpus: Corpus initialized from prodigy datasets
        """
        train_ds = Dataset("train").from_prodigy(prodigy_train_datasets)
        dev_ds = Dataset("dev").from_prodigy(prodigy_dev_datasets)
        test_ds = Dataset("test").from_prodigy(prodigy_test_datasets) if prodigy_test_datasets else None

        ds = cls(name, train_ds, dev_ds, test_ds)
        return ds

    def to_prodigy(
        self,
        name: str = None,
        prodigy_train_dataset: str = None,
        prodigy_dev_dataset: str = None,
        prodigy_test_dataset: str = None,
        overwrite: bool = True,
    ) -> Tuple[str, str, str]:
        """Save a Corpus to 3 separate Prodigy datasets

        Args:
            prodigy_train_dataset (str, optional): Train dataset name in Prodigy
            prodigy_dev_dataset (str, optional): Dev dataset name in Prodigy
            prodigy_test_dataset (str, optional): Test dataset name in Prodigy
        """
        name = name if name else self.name

        if not prodigy_train_dataset:
            prodigy_train_dataset = f"{name}_train_{self.train_ds.commit_hash}"

        if not prodigy_dev_dataset:
            prodigy_dev_dataset = f"{name}_dev_{self.dev_ds.commit_hash}"

        if not prodigy_test_dataset:
            prodigy_test_dataset = f"{name}_test_{self.test_ds.commit_hash}"

        self.train_ds.to_prodigy(prodigy_train_dataset, overwrite=overwrite)
        self.dev_ds.to_prodigy(prodigy_dev_dataset, overwrite=overwrite)
        self.test_ds.to_prodigy(prodigy_test_dataset, overwrite=overwrite)

        return (prodigy_train_dataset, prodigy_dev_dataset, prodigy_test_dataset)



import catalogue


class registry:
    operations = catalogue.create("recon", "operations", entry_points=True)
    operation_factories = catalogue.create("recon", "operation_factories", entry_points=True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['recon', 'recon.cli', 'recon.operations', 'recon.prodigy']

package_data = \
{'': ['*'], 'recon.prodigy': ['templates/*']}

install_requires = \
['numpy>=1.22.0,<1.23.0',
 'scipy>=1.8.0,<1.9',
 'spacy>=3.2.0,<3.4.0',
 'xxhash>=3.0.0,<3.1.0']

entry_points = \
{'console_scripts': ['recon = recon.cli:app'],
 'prodigy_recipes': ['recon.ner_correct = recon:prodigy_recipes.ner_correct',
                     'recon.ner_merge = recon:prodigy_recipes.ner_merge']}

setup_kwargs = {
    'name': 'reconner',
    'version': '0.9.1',
    'description': 'Recon NER, Debug and correct annotated Named Entity Recognition (NER) data for inconsitencies and get insights on improving the quality of your data.',
    'long_description': None,
    'author': 'Kabir Khan',
    'author_email': 'kabirkhan1137@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://kabirkhan.github.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import math
from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Sequence, Union, cast

import numpy as np
import srsly
from recon.constants import NOT_LABELED
from recon.types import EntityCoverage, EntityCoverageStats, Example, NERStats, Outliers
from scipy.spatial.distance import jensenshannon
from scipy.stats import entropy as scipy_entropy


def get_ner_stats(
    data: List[Example], serialize: bool = False, return_examples: bool = False
) -> Union[NERStats, str, None]:
    """Compute statistics for NER data

    Args:
        data (Iterator[Example]): Data as a List of examples
        serialize (bool, optional): Serialize to a JSON string for printing.
        return_examples (bool, optional): Whether to return examples per type

    Returns:
        Union[NERStats, str, None]:
            List of examples or string if serialize and no_print are both True
    """
    annotations_per_type: DefaultDict[str, Any] = defaultdict(int)
    examples: DefaultDict[str, Any] = defaultdict(list)
    n_examples_no_entities = 0

    length = 0

    for e in data:
        length += 1
        if not e.spans:
            n_examples_no_entities += 1
            examples[NOT_LABELED].append(e)
        else:
            for s in e.spans:
                annotations_per_type[s.label] += 1
                examples[s.label].append(e)

    sorted_anns_by_count = {a[0]: a[1] for a in sorted(annotations_per_type.items(), key=lambda x: x[1], reverse=True)}

    stats = NERStats(
        n_examples=length,
        n_examples_no_entities=n_examples_no_entities,
        n_annotations=sum(annotations_per_type.values()),
        n_annotations_per_type=sorted_anns_by_count,
    )
    if return_examples:
        stats.examples_with_type = examples

    if serialize:
        return srsly.json_dumps(stats.dict(), indent=4)
    else:
        return stats


def get_sorted_type_counts(ner_stats: NERStats) -> List[int]:
    """Get list of counts for each type in n_annotations_per_type property
    of an NERStats object sorted by type name

    Args:
        ner_stats (NERStats): Dataset stats

    Returns:
        List[int]: List of counts sorted by type name
    """
    annotations_per_type = ner_stats.n_annotations_per_type
    annotations_per_type[NOT_LABELED] = ner_stats.n_examples_no_entities

    return [t[1] for t in sorted(annotations_per_type.items(), key=lambda p: p[0])]


def calculate_label_distribution_similarity(x: List[Example], y: List[Example]) -> float:
    """Calculate the similarity of the label distribution for 2 datasets.

    e.g. This can help you understand how well your train set models your dev and test sets.
    Empircally you want a similarity over **0.8** when comparing your train set to each of your
    dev and test sets.

        calculate_label_distribution_similarity(corpus.train, corpus.dev)
        # 98.57

        calculate_label_distribution_similarity(corpus.train, corpus.test)
        # 73.29 - This is bad, let's investigate our test set more

    Args:
        x (List[Example]): Dataset
        y (List[Example]): Dataset to compare x to

    Returns:
        float: Similarity of label distributions
    """

    def pipeline(data: List[Example]) -> Sequence[float]:
        stats = cast(NERStats, get_ner_stats(data))
        sorted_type_counts = get_sorted_type_counts(stats)
        counts_to_probs = get_probs_from_counts(sorted_type_counts)
        return counts_to_probs

    distance = jensenshannon(pipeline(x), pipeline(y))

    return (1 - distance) * 100


def get_entity_coverage(
    data: List[Example],
    sep: str = "||",
    case_sensitive: bool = False,
    return_examples: bool = False,
) -> List[EntityCoverage]:
    """Identify how well you dataset covers an entity type. Get insights
    on the how many times certain text/label span combinations exist across your
    data so that you can focus your annotation efforts better rather than
    annotating examples your Model already understands well.

    Args:
        data (List[Example]): List of examples
        sep (str, optional): Separator used in coverage map, only change if || exists in your text
            or label.
        case_sensitive (bool, optional): Consider case of text for each annotation
        return_examples (bool, optional): Return Examples that contain the entity label annotation.

    Returns:
        List[EntityCoverage]: Sorted List of EntityCoverage objects containing the text, label, count, and
            an optional list of examples where that text/label annotation exists.
    """
    coverage_map: DefaultDict[str, int] = defaultdict(int)
    examples_map: DefaultDict[str, List[Example]] = defaultdict(list)

    for example in data:
        for span in example.spans:
            text = span.text if case_sensitive else span.text.lower()
            key = f"{text}{sep}{span.label}"
            coverage_map[key] += 1
            examples_map[key].append(example)

    coverage = []
    for key, count in coverage_map.items():
        text, label = key.split(sep)
        record = EntityCoverage(text=text, label=label, count=count)
        if return_examples:
            record.examples = examples_map[key]
        coverage.append(record)

    sorted_coverage = sorted(coverage, key=lambda x: x.count, reverse=True)
    return sorted_coverage


def calculate_entity_coverage_similarity(x: List[Example], y: List[Example]) -> EntityCoverageStats:
    """Calculate how well dataset x covers the entities in dataset y.
    This function should be used to calculate how similar your train set
    annotations cover the annotations in your dev/test set

    Args:
        x (List[Example]): Dataset to compare coverage to (usually corpus.train)
        y (List[Example]): Dataset to evaluate coverage for (usually corpus.dev or corpus.test)

    Returns:
        EntityCoverageStats: Stats with
            1. The base entity coverage (does entity in y exist in x)
            2. Count coverage (sum of the EntityCoverage.count property for
            each EntityCoverage in y to get a more holisic coverage scaled by how
            often entities occur in each dataset x and y)
    """

    def pipeline(data: List[Example]) -> Dict[int, int]:
        ecs = get_entity_coverage(data)
        return {hash(ec): ec.count for ec in ecs}

    x_map = pipeline(x)
    y_map = pipeline(y)

    n_intersection = 0
    count_intersection = 0
    n_union = 0
    count_union = 0

    for k, count in y_map.items():
        if k in x_map:
            n_intersection += 1
            count_intersection += count
        n_union += 1
        count_union += count

    return EntityCoverageStats(
        entity=(n_intersection / n_union) * 100,
        count=(count_intersection / count_union) * 100,
    )


def get_probs_from_counts(seq: Sequence[int]) -> Sequence[float]:
    """Convert a sequence of counts to a sequence of probabilties
    by dividing each n by the sum of all n in seq

    Args:
        seq (Sequence[int]): Sequence of counts

    Returns:
        Sequence[float]: Sequence of probabilities
    """
    return np.asarray(seq) / sum(seq)


def entropy(seq: Union[List[int], List[float]], total: int = None) -> float:
    """Calculate Shannon Entropy for a sequence of Floats or Integers.
    If Floats, check they are probabilities
    If Integers, divide each n in seq by total and calculate entropy

    Args:
        seq (Union[List[int], List[float]]): Sequence to calculate entropy for
        total (int, optional): Total to divide by for List of int

    Raises:
        ValueError: If seq is not valid

    Returns:
        float: Entropy for sequence
    """
    if not seq:
        raise ValueError("Pass a valid non-empty sequence")

    if isinstance(seq[0], float):
        e = scipy_entropy(seq)
    elif isinstance(seq[0], int):
        e = scipy_entropy(get_probs_from_counts(seq))
    else:
        raise ValueError("Parameter seq must be a sequence of probabilites or integers.")
    return e


def calculate_label_balance_entropy(ner_stats: NERStats) -> float:
    """Use Entropy to calculate a metric for label balance based on an NERStats object

    Args:
        ner_stats (NERStats): NERStats for a dataset.

    Returns:
        float: Entropy for annotation counts of each label
    """
    total = ner_stats.n_annotations
    classes = [count for label, count in ner_stats.n_annotations_per_type.items()]
    return entropy(classes, total)


def calculate_entity_coverage_entropy(
    entity_coverage: List[EntityCoverage],
) -> float:
    """Use Entropy to calculate a metric for entity coverage.

    Args:
        entity_coverage (List[EntityCoverage]): List of EntityCoverage
            from get_entity_coverage

    Returns:
        float: Entropy for entity coverage counts
    """
    counts = [ecs.count for ecs in entity_coverage]
    return entropy(counts, sum(counts))  # type: ignore


def detect_outliers(seq: Sequence[Any], use_log: bool = False) -> Outliers:
    """Detect outliers in a numerical sequence.

    Args:
        seq (Sequence[Any]): Sequence of ints or floats
        use_log (bool, optional): Use logarithm of seq.

    Returns:
        Tuple[List[int], List[int]]: Tuple of low and high indices
    """
    q1 = np.quantile(seq, 0.25)
    q3 = np.quantile(seq, 0.75)
    iqr = q3 - q1
    fence_low = math.floor(q1 - 1.5 * iqr)
    fence_high = math.floor(q3 + 1.5 * iqr)
    low_indices = [i for i, n in enumerate(seq) if n <= fence_low]
    high_indices = [i for i, n in enumerate(seq) if n > fence_high]
    return Outliers(low=low_indices, high=high_indices)



from typing import List

from recon.operations.core import operation
from recon.types import Example, Span


@operation("recon.v1.upcase_labels")
def upcase_labels(example: Example) -> Example:
    """Convert all span labels to uppercase to normalize

    Args:
        example (Example): Input Example

    Returns:
        Example: Example with fixed labels
    """
    for s in example.spans:
        s.label = s.label.upper()
    return example


@operation("recon.v1.filter_overlaps")
def filter_overlaps(example: Example) -> Example:
    """Filter overlapping entity spans by picking the longest one.

    Args:
        example (Example): Input Example

    Returns:
        List[Example]: Example with fixed overlaps
    """
    annotations: List[Span] = sorted(example.spans, key=lambda s: s.start)
    filtered_annotations = remove_overlapping_entities(annotations)
    example.spans = filtered_annotations

    return example


def select_subset_of_overlapping_chain(chain: List[Span]) -> List[Span]:
    """
    Select the subset of entities in an overlapping chain to return by greedily choosing the
    longest entity in the chain until there are no entities remaining
    """
    sorted_chain = sorted(chain, key=lambda s: s.end - s.start, reverse=True)
    selections_from_chain: List[Span] = []
    chain_index = 0
    # dump the current chain by greedily keeping the longest entity that doesn't overlap
    while chain_index < len(sorted_chain):
        entity = sorted_chain[chain_index]
        match_found = False
        for already_selected_entity in selections_from_chain:
            max_start = max(entity.start, already_selected_entity.start)
            min_end = min(entity.end, already_selected_entity.end)
            if len(range(max_start, min_end)) > 0:
                match_found = True
                break

        if not match_found:
            selections_from_chain.append(entity)

        chain_index += 1

    return selections_from_chain


def remove_overlapping_entities(sorted_spans: List[Span]) -> List[Span]:
    """
    Removes overlapping entities from the entity set, by greedilytaking the longest
    entity from each overlapping chain. The input list of entities should be sorted
    and follow the spacy format.
    """
    spans_without_overlap: List[Span] = []
    current_overlapping_chain: List[Span] = []
    current_overlapping_chain_start = 0
    current_overlapping_chain_end = 0
    for i, current_entity in enumerate(sorted_spans):
        current_entity = sorted_spans[i]
        current_entity_start = current_entity.start
        current_entity_end = current_entity.end

        if len(current_overlapping_chain) == 0:
            current_overlapping_chain.append(current_entity)
            current_overlapping_chain_start = current_entity_start
            current_overlapping_chain_end = current_entity_end
        else:
            min_end = min(current_entity_end, current_overlapping_chain_end)
            max_start = max(current_entity_start, current_overlapping_chain_start)
            if min_end - max_start > 0:
                current_overlapping_chain.append(current_entity)
                current_overlapping_chain_start = min(current_entity_start, current_overlapping_chain_start)
                current_overlapping_chain_end = max(current_entity_end, current_overlapping_chain_end)
            else:
                selections_from_chain: List[Span] = select_subset_of_overlapping_chain(current_overlapping_chain)

                current_overlapping_chain = []
                spans_without_overlap.extend(selections_from_chain)
                current_overlapping_chain.append(current_entity)
                current_overlapping_chain_start = current_entity_start
                current_overlapping_chain_end = current_entity_end

    spans_without_overlap.extend(select_subset_of_overlapping_chain(current_overlapping_chain))

    return sorted(spans_without_overlap, key=lambda x: x.start)



"""Load a list of Example data from JSON represented records in 
the [Prodigy](https://prodi.gy) format.
"""

from pathlib import Path
from typing import Any, Dict, Iterable, List

import spacy
import srsly
from recon.types import Example, Span, Token
from spacy.language import Language
from spacy.tokens import Doc, DocBin
from spacy.util import get_words_and_spaces


def read_jsonl(path: Path) -> List[Example]:
    """Read annotations in JSONL file format

    Args:
        path (Path): Path to data

    Returns:
        List[Example]: List of examples
    """
    data = srsly.read_jsonl(path)
    examples = json_to_examples(data)
    return examples


def read_json(path: Path) -> List[Example]:
    """Read annotations in JSON file format

    Args:
        path (Path): Path to data

    Returns:
        List[Example]: List of examples
    """
    data = srsly.read_jsonl(path)
    examples = json_to_examples(data)
    return examples


def json_to_examples(data: List[Dict[str, Any]]) -> List[Example]:
    """Convert List of Dicts to List of typed Examples

    Args:
        data (List[Dict[str, Any]]): Input List of Dicts to convert

    Returns:
        List[Example]: List of typed Examples
    """
    return [Example(**example) for example in data]


def from_spacy(path: Path, nlp: Language = None, lang_code: str = "en") -> Iterable[Example]:
    """Load examples from .spacy docbin format

    Args:
        path (Path): Path to data
        nlp (Language, optional): Spacy Language object.
        lang_code (str, optional): Language code to create a blank spacy model with if nlp is not provided.

    Yields:
        Iterable[Example]: List of typed Examples
    """
    if not nlp:
        nlp = spacy.blank(lang_code)

    doc_bin = DocBin().from_disk(path)
    for doc in doc_bin.get_docs(nlp.vocab):
        yield Example(
            text=doc.text,
            spans=[
                Span(
                    text=e.text,
                    start=e.start_char,
                    end=e.end_char,
                    label=e.label_,
                    token_start=e.start,
                    token_end=e.end,
                )
                for e in doc.ents
            ],
            tokens=[Token(text=t.text, start=t.idx, end=t.idx + len(t), id=t.i) for t in doc],
        )


def to_spacy(path: Path, data: Iterable[Example], nlp: Language = None, lang_code: str = "en") -> DocBin:
    """Save a batch of examples to disk in the .spacy DocBin format

    Args:
        path (Path): Path to data
        data (Iterable[Example]): Input Examples
        nlp (Language, optional): Spacy Language object.
        lang_code (str, optional): Language code to create a blank spacy model with if nlp is not provided.

    Returns:
        DocBin: Spacy DocBin with stored example data.
    """

    if not nlp:
        nlp = spacy.blank(lang_code)

    doc_bin = DocBin(attrs=["ENT_IOB", "ENT_TYPE"])
    for example in data:
        if example.tokens:
            tokens = [token.text for token in example.tokens]
            words, spaces = get_words_and_spaces(tokens, example.text)
            doc = Doc(nlp.vocab, words=words, spaces=spaces)
            doc.ents = [doc.char_span(s.start, s.end, label=s.label) for s in example.spans]
            doc_bin.add(doc)
    doc_bin.to_disk(path)
    return doc_bin



import warnings
from collections import Counter, defaultdict
from typing import Any, Callable, Dict, Iterator, List, Tuple, Union

from pydantic.error_wrappers import ErrorWrapper
from recon.operations import registry as op_registry
from recon.operations.utils import (
    get_received_operation_data,
    get_required_operation_params,
    request_body_to_args,
)
from recon.preprocess import PreProcessor
from recon.preprocess import registry as pre_registry
from recon.types import (
    Example,
    OperationResult,
    OperationState,
    OperationStatus,
    Transformation,
    TransformationType,
)
from tqdm import tqdm
from wasabi import Printer


def op_iter(
    data: List[Example], pre: List[PreProcessor], verbose: bool = True
) -> Iterator[Tuple[int, Example, Dict[str, Any]]]:
    """Iterate over list of examples for an operation
    yielding tuples of (example hash, example)

    Args:
        data (List[Example]): List of examples to iterate
        pre (List[PreProcessor]): List of preprocessors to run
        verbose (bool, optional): Show verbose output.

    Yields:
        Iterator[Tuple[int, Example]]: Tuples of (example hash, example)
    """
    msg = Printer(no_print=verbose == False, hide_animation=verbose == False)
    preprocessed_outputs: Dict[Example, Dict[str, Any]] = defaultdict(dict)
    for processor in pre:
        msg.info(f"\t=> Running preprocessor {processor.name}")
        processor_outputs = processor(data)
        for example, output in tqdm(zip(data, processor_outputs), total=len(data), disable=(not verbose), leave=False):
            preprocessed_outputs[example][processor.name] = output
            example.__setattr__(processor.field, output)

    for example in data:
        yield hash(example), example.copy(deep=True), preprocessed_outputs[example]


class operation:
    def __init__(
        self,
        name: str,
        pre: List[Union[str, PreProcessor]] = [],
        handles_tokens: bool = True,
        factory: bool = False,
        augmentation: bool = False,
    ):
        """Decorate an operation that makes some changes to a dataset.

        Args:
            name (str): Operation name.
            pre (Union[List[str], List[PreProcessor]]): List of preprocessors to run
        """
        self.name = name
        self.pre = pre
        self.handles_tokens = handles_tokens
        self.factory = factory
        self.augmentation = augmentation

    def __call__(self, *args: Any, **kwargs: Any) -> Callable:
        """Decorator for an operation.
        The first arg is the function being decorated.
        This function can either operate on a List[Example]
        and in that case self.batch should be True.

        e.g. @operation("recon.v1.some_name", batch=True)

        Or it should operate on a single example and
        recon will take care of applying it to a full Dataset

        Args:
            args: First arg is function to decorate

        Returns:
            Callable: Original function
        """
        op: Callable = args[0]

        pre: List[PreProcessor] = []

        for pre_name_or_op in self.pre:
            preprocessor = pre_name_or_op
            if isinstance(preprocessor, str):
                preprocessor = pre_registry.preprocessors.get(pre_name_or_op)
            assert isinstance(preprocessor, PreProcessor)
            pre.append(preprocessor)

        if self.factory:

            def factory(pre: List[PreProcessor]) -> Operation:
                return Operation(
                    self.name,
                    pre,
                    op=op,
                    handles_tokens=self.handles_tokens,
                    augmentation=self.augmentation,
                )

            op_registry.operation_factories.register(self.name)(factory)
        else:
            op_registry.operations.register(self.name)(
                Operation(self.name, pre, op, self.handles_tokens, augmentation=self.augmentation)
            )

        return op


class Operation:
    """Operation class that takes care of calling and reporting
    the results of an operation on a Dataset"""

    def __init__(
        self,
        name: str,
        pre: List[PreProcessor],
        op: Callable,
        handles_tokens: bool,
        augmentation: bool,
    ):
        """Initialize an Operation instance

        Args:
            name (str): Name of operation
            pre (List[PreProcessor]): List of preprocessors to run
            op (Callable): Decorated function
        """
        self.name = name
        self.pre = pre
        self.op = op
        self.handles_tokens = handles_tokens
        self.augmentation = augmentation

    def __call__(self, dataset: Any, *args: Any, **kwargs: Any) -> OperationResult:
        """Runs op on a dataset and records the results

        Args:
            dataset (Dataset): Dataset to operate on

        Raises:
            ValueError: if track_example is called in the op with no data

        Returns:
            OperationResult: Container holding new data and the state of the Operation
        """
        verbose = kwargs.pop("verbose", False)
        initial_state = kwargs.pop("initial_state", None)
        if not initial_state:
            initial_state = OperationState(name=self.name)
        state = initial_state.copy(deep=True)

        if state.status == OperationStatus.NOT_STARTED:
            state.status = OperationStatus.IN_PROGRESS

        state.args = args
        state.kwargs = kwargs

        def add_example(new_example: Example) -> None:
            state.transformations.append(
                Transformation(example=hash(new_example), type=TransformationType.EXAMPLE_ADDED)
            )
            dataset.example_store.add(new_example)

        def remove_example(orig_example_hash: int) -> None:
            state.transformations.append(
                Transformation(prev_example=orig_example_hash, type=TransformationType.EXAMPLE_REMOVED)
            )

        def change_example(orig_example_hash: int, new_example: Example) -> None:
            state.transformations.append(
                Transformation(
                    prev_example=orig_example_hash,
                    example=hash(new_example),
                    type=TransformationType.EXAMPLE_CHANGED,
                )
            )
            dataset.example_store.add(new_example)

        has_tokens = False
        for e in dataset.data:
            if e.tokens or any([(s.token_start or s.token_end) for s in e.spans]):
                has_tokens = True
                break

        if has_tokens and not self.handles_tokens:
            warnings.warn(
                # fmt: off
                "This dataset seems to have preset tokens. " +
                f"Operation: {self.name} is not currently capable of handling tokens and you will "
                "need to reset tokenization after this operation. " +
                "Applying the `recon.v1.add_tokens` operation after this " +
                "is complete will get you back to a clean state."
                # fmt: on
            )

        required_params = get_required_operation_params(self.op)
        received_data = get_received_operation_data(required_params, state)

        values: Dict[str, Any] = {}
        errors: List[ErrorWrapper] = []

        if received_data:
            values, errors = request_body_to_args(list(required_params.values()), received_data)
            print("VALUES: ", values)

        if errors:
            error_msg = (
                f"Validation error while trying to call operation: {self.name} "
                + "with provided args and kwargs values. "
            )
            for err in errors:
                error_msg += str(err.exc)
                print(values)
            raise ValueError(error_msg)

        state.args = ()
        state.kwargs = values
        new_data = []

        with tqdm(total=len(dataset), disable=(not verbose)) as pbar:
            for orig_example_hash, example, preprocessed_outputs in op_iter(dataset.data, self.pre, verbose=verbose):
                if preprocessed_outputs:
                    res = self.op(example, preprocessed_outputs=preprocessed_outputs, **values)
                else:
                    res = self.op(example, **values)

                if res is None:
                    remove_example(orig_example_hash)
                elif isinstance(res, list):
                    old_example_present = False
                    for new_example in res:
                        new_data.append(new_example)
                        if hash(new_example) == orig_example_hash:
                            old_example_present = True
                        else:
                            add_example(new_example)
                    if not old_example_present:
                        remove_example(orig_example_hash)
                else:
                    assert isinstance(res, Example)
                    new_data.append(res)
                    if hash(res) != orig_example_hash:
                        change_example(orig_example_hash, res)

                pbar.update(1)

        transformation_counts = Counter([t.type for t in state.transformations])

        state.examples_added = transformation_counts[TransformationType.EXAMPLE_ADDED]
        state.examples_removed = transformation_counts[TransformationType.EXAMPLE_REMOVED]
        state.examples_changed = transformation_counts[TransformationType.EXAMPLE_CHANGED]
        state.status = OperationStatus.COMPLETED

        state_copy = state.copy(deep=True)
        state = OperationState(name=self.name)
        return OperationResult(data=new_data, state=state_copy)

    def register(self) -> None:
        op_registry.operations.register(self.name)(self)



from typing import Any, Callable, Dict, List, Optional

import numpy as np
from recon.operations.core import operation
from recon.types import Example, Span


def mask_1d(length: int, prob: float = 0.5) -> np.ndarray:
    if prob < 0 or prob > 1:
        raise ValueError(f"Prob of {prob} is not allowed. Allowed values between 0 and 1.")

    mask = np.zeros(length, dtype=int)
    mask[: np.ceil(length * prob).astype(int)] = 1
    np.random.shuffle(mask)
    mask = mask.astype(bool)
    return mask


def substitute_spans(example: Example, span_subs: Dict[Span, str]) -> Example:
    """Substitute spans in an example. Replaces span text and alters the example text
    and span offsets to create a valid example.

    Args:
        example (Example): Input example
        span_subs (Dict[int, str]): Mapping of span hash to a str replacement text

    Returns:
        Example: Output example with substituted spans
    """
    span_sub_start_counter = 0

    new_example_text = example.text
    new_example_spans = []

    prev_example_spans = {hash(span) for span in example.spans}
    spans = sorted(set(list(span_subs.keys()) + example.spans), key=lambda s: s.start)

    for span in spans:
        should_add_span = hash(span) in prev_example_spans

        prev_end = span.end
        new_text = span.text

        if span in span_subs:
            new_text = span_subs[span]
            new_start = span.start + span_sub_start_counter
            new_end = new_start + len(new_text)

            new_example_text = (
                new_example_text[: span.start + span_sub_start_counter]
                + new_text
                + new_example_text[span.end + span_sub_start_counter :]
            )

            span.text = new_text
            span.start = new_start
            span.end = new_end

            span_sub_start_counter += new_end - prev_end
        else:
            span.start += span_sub_start_counter
            span.end = span.start + len(new_text)
            span_sub_start_counter = span.end - prev_end

        span.text = new_text

        if should_add_span:
            new_example_spans.append(span)

    example.text = new_example_text
    example.spans = new_example_spans

    return example


def augment_example(
    example: Example,
    span_f: Callable[[Span, Any], Optional[str]],
    spans: List[Span] = None,
    span_label: str = None,
    n_augs: int = 1,
    sub_prob: float = 0.5,
    **kwargs: Any,
) -> List[Example]:

    if spans is None:
        spans = example.copy(deep=True).spans

    augmented_examples = {example}

    for i in range(n_augs):
        example = example.copy(deep=True)
        if span_label:
            spans = [s for s in spans if s.label == span_label]
        mask = mask_1d(len(spans), prob=sub_prob)
        spans_to_sub = list(np.asarray(spans)[mask])

        span_subs = {}
        for span in spans_to_sub:
            res = span_f(span, **kwargs)  #  type: ignore
            if res:
                span_subs[span] = res

        if not any(span_subs.values()) or len(augmented_examples) > n_augs:
            break

        example = substitute_spans(example, span_subs)
        if example not in augmented_examples:
            augmented_examples.add(example)

    return list(augmented_examples)


@operation("recon.v1.augment.ent_label_sub", handles_tokens=False)
def ent_label_sub(
    example: Example, label: str, subs: List[str], n_augs: int = 1, sub_prob: float = 0.5
) -> List[Example]:
    """Augmentation to substitute entities based on label.
    Applies a mask to the entities that have a provided label based on substitution_prob
    and selects a random choice from the list of provided substitutions to replace each
    span with

    Args:
        example (Example): Input example
        label (str): Span label to replace
            e.g. PERSON or LOCATION
        subs (List[str]): List of substitutions
            e.g. list of names if label is PERSON
        n_augs (int, optional): Maximum number of augmentated examples to create per example.
        sub_prob (float, optional): Probability from 0-1 inclusive of how many of the spans to replace

    Returns:
        List[Example]: List of augmented examples including the original.
    """

    def augmentation(span: Span, subs: List[str]) -> Optional[str]:
        subs = [s for s in subs if s != span.text]
        sub = None
        if len(subs) > 0:
            sub = np.random.choice(subs)
        return sub

    return augment_example(example, augmentation, n_augs=n_augs, sub_prob=sub_prob, subs=subs)


@operation("recon.v1.augment.kb_expansion", factory=True)
def kb_expansion(
    example: Example,
    preprocessed_outputs: Dict[str, Any] = {},
    n_augs: int = 1,
    sub_prob: float = 0.5,
) -> List[Example]:

    spans_to_aliases_map = preprocessed_outputs["recon.v1.span_aliases"]

    def augmentation(span: Span, spans_to_aliases_map: Dict[int, List[str]]) -> Optional[str]:
        sub = None
        if hash(span) in spans_to_aliases_map:
            aliases = spans_to_aliases_map[hash(span)]

            if len(aliases) > 0:
                rand_alias = np.random.choice(aliases)
                index = aliases.index(rand_alias)
                del spans_to_aliases_map[hash(span)][index]
                sub = rand_alias

        return sub

    return augment_example(
        example,
        augmentation,
        n_augs=n_augs,
        sub_prob=sub_prob,
        spans_to_aliases_map=spans_to_aliases_map,
    )


# def get_synonym(word, pos=None):
#     """Get synonym for word given its part-of-speech (pos)."""
#     synsets = wn.synsets(word, pos=pos)
#     # Return None if wordnet has no synsets (synonym sets) for this word and pos.
#     if synsets:
#         print(synsets[0])
#         words = [lemma.name() for lemma in synsets[0].lemmas()]
#         if words[0].lower() != word.lower():  # Skip if synonym is same as word.
#             # Multi word synonyms in wordnet use '_' as a separator e.g. reckon_with. Replace it with space.
#             return words[0].replace("_", " ")


# @operation("recon.v1.augment.replace_pos_with_synonym", pre=[spacy_pre])
# def replace_pos_with_synonym(example: Example, pos: str, synonym_f: Callable[[str], str] = get_synonym, preprocessed_outputs={}, n_augs: int = 1):

#     pos_map = {
#         "VERB": "v",
#         "NOUN": "n",
#         "ADJ": "a"
#     }

#     if pos not in pos_map:
#         raise ValueError(f"Argument `pos` of {pos} not in {''.join(pos_map.keys())}")

#     doc = example.data.doc
#     span_starts = [s.start for s in example.spans]

#     # Get indices of verb tokens in sentence.
#     pos_idxs = [i for i, token in enumerate(doc) if token.pos_ == pos and token.idx not in span_starts]
#     tokens = [doc[idx] for idx in pos_idxs]
#     spans = [Span(text=token.text, start=token.idx, end=token.idx + len(token.text), label="") for token in tokens]

#     def augmentation(span: Span, synonym_f: Callable[[str], str] = synonym_f) -> Optional[str]:
#         return synonym_f(span.text)

#     return augment_example(
#         example,
#         augmentation,
#         spans=spans,
#         n_augs=n_augs,
#         sub_prob=sub_prob,
#         spans_to_aliases_map=spans_to_aliases_map,
#     )

#     for i in range(n_augs):
#         example = example.copy(deep=True)

#         if pos_idxs:
#             # Pick random verb idx to replace.
#             idx = np.random.choice(pos_idxs)
#             token = doc[idx]
#             synonym = get_synonyms(token.text, pos="v")
#             synonym = "enjoy"
#             print(synonym)


#             # If there's a valid verb synonym, replace it. Otherwise, return None.
#             if synonym:
#                 curr_span = Span(text=token.text, start=token.idx, end=token.idx + len(token.text), label="")
#                 span_subs = {curr_span: synonym}
#                 example = substitute_spans(example, span_subs)
#                 if example not in augmented_examples:
#                     augmented_examples.add(example)

#     return list(augmented_examples)



"""Make corrections to your data."""

from typing import Any, Dict, List, cast

from recon.operations.core import operation
from recon.types import Correction, Example, Span, Token
from spacy.tokens import Span as SpacySpan
from wasabi import msg


@operation("recon.v1.rename_labels")
def rename_labels(example: Example, label_map: Dict[str, str]) -> Example:
    """Rename labels in a copy of List[Example] data

    Args:
        example (Example): Input Example
        label_map (Dict[str, str]): One-to-one mapping of label names

    Returns:
        Example: Copy of Example with renamed labels
    """
    for span in example.spans:
        span.label = label_map.get(span.label, span.label)
    return example


@operation("recon.v1.fix_annotations")
def fix_annotations(
    example: Example,
    corrections: List[Correction],
    case_sensitive: bool = False,
    dryrun: bool = False,
) -> Example:
    """Fix annotations in a copy of List[Example] data.

    This function will NOT add annotations to your data.
    It will only remove erroneous annotations and fix the
    labels for specific spans.

    Args:
        example (Example): Input Example
        corrections (Dict[str, str]): Dictionary of corrections mapping entity text to a new label.
            If the value is set to None, the annotation will be removed
        case_sensitive (bool, optional): Consider case of text for each correction
        dryrun (bool, optional): Treat corrections as a dryrun and just print all changes to be made

    Returns:
        Example: Example with fixed annotations
    """

    if not case_sensitive:
        for c in corrections:
            c.annotation = c.annotation.lower()

    corrections_map: Dict[str, Correction] = {c.annotation: c for c in corrections}
    prints: List[str] = []

    ents_to_remove: List[int] = []
    for i, s in enumerate(example.spans):
        t = s.text if case_sensitive else s.text.lower()

        if t in corrections_map:
            c = corrections_map[t]
            if c.to_label is None and (s.label in c.from_labels or "ANY" in c.from_labels):
                if dryrun:
                    prints.append(f"Deleting span: {s.text}")
                else:
                    ents_to_remove.append(i)
            elif s.label in c.from_labels or "ANY" in c.from_labels:
                if dryrun:
                    prints.append(f"Correction span: {s.text} from labels: {c.from_labels} to label: {c.to_label}")
                else:
                    s.label = cast(str, c.to_label)

    i = len(ents_to_remove) - 1
    while i >= 0:
        idx = ents_to_remove[i]
        del example.spans[idx]
        i -= 1

    if dryrun:
        msg.divider("Example Text")
        msg.text(example.text)
        for line in prints:
            msg.text(line)

    return example


def corrections_from_dict(corrections_dict: Dict[str, Any]) -> List[Correction]:
    """Create a list of Correction objects from a simpler config for
    corrections using a Dict representation mapping keys to either the label to
    convert to or a tuple of (from_label, to_label) pairings or (List[from_labels], to_label)
    pairings if you want to convert as subset of labels at a time

    Args:
        corrections_dict (Dict[str, Any]): Corrections formatted dict
            e.g. {
                "united states": "GPE",
                "London": (["LOC"], "GPE")
            }

    Raises:
        ValueError: If the format of the dict

    Returns:
        [type]: [description]
    """
    corrections: List[Correction] = []
    for key, val in corrections_dict.items():
        if isinstance(val, str) or val == None:
            from_labels = ["ANY"]
            to_label = val
        elif isinstance(val, tuple):
            if isinstance(val[0], str):
                from_labels = [val[0]]
            else:
                from_labels = val[0]
            to_label = val[1]
        else:
            raise ValueError(
                "Cannot parse corrections dict. Value must be either a str of the label "
                + "to change the annotation to (TO_LABEL) or a tuple of (FROM_LABEL, TO_LABEL)"
            )
        corrections.append(Correction(annotation=key, from_labels=from_labels, to_label=to_label))
    return corrections


@operation("recon.v1.strip_annotations", pre=["recon.v1.spacy"], handles_tokens=False)
def strip_annotations(
    example: Example,
    strip_chars: List[str] = [".", "!", "?", "-", ":", " "],
    preprocessed_outputs: Dict[str, Any] = {},
) -> Example:
    """Strip punctuation and spaces from start and end of annotations.
    These characters are almost always a mistake and will confuse a model

    Args:
        example (Example): Input Example
        strip_chars (List[str], optional): Characters to strip.

    Returns:
        Example: Example with stripped spans
    """

    preprocessed_outputs["recon.v1.spacy"]

    def fix_tokens(example: Example, span: Span) -> bool:
        fix_tokens = example.tokens and any(example.tokens)
        return bool(fix_tokens and span.token_start and span.token_end and span.token_start < span.token_end)

    for s in example.spans:
        for ch in strip_chars:
            if s.text.startswith(ch):
                ch = s.text[0]
                while ch in strip_chars:
                    s.text = s.text[1:]
                    s.start += 1
                    ch = s.text[0]
            elif s.text.endswith(ch):
                ch = s.text[-1]
                while ch in strip_chars:
                    s.text = s.text[:-1]
                    ch = s.text[-1]
                    s.end -= 1
    return example


@operation("recon.v1.split_sentences", pre=["recon.v1.spacy"])
def split_sentences(example: Example, preprocessed_outputs: Dict[str, Any] = {}) -> List[Example]:
    """Split a single example into multiple examples by splitting the text into
    multiple sentences and resetting entity and token offsets based on offsets
    relative to sentence boundaries

    Args:
        example (Example): Input Example
        preprocessed_outputs (Dict[str, Any], optional): Outputs of preprocessors.

    Returns:
        List[Example]: List of split examples.
            Could be list of 1 if the example is just one sentence.
    """
    doc = preprocessed_outputs["recon.v1.spacy"]

    new_examples = []
    ents = []
    for ent in example.spans:
        span = doc.char_span(ent.start, ent.end, label=ent.label)
        if not span:
            token = None
            text = doc.text[ent.start : ent.end]
            for t in doc:
                if t.text == text:
                    token = t
            if token:
                span = SpacySpan(doc, token.i, token.i + 1, label=ent.label)
        ents.append(span)

    doc.ents = ents

    for sent in doc.sents:
        sent_doc = sent.as_doc()
        new_example = Example(
            text=sent.text,
            spans=[
                Span(
                    text=e.text,
                    start=e.start_char,
                    end=e.end_char,
                    token_start=e.start,
                    token_end=e.end,
                    label=e.label_,
                )
                for e in sent_doc.ents
            ],
            tokens=[Token(text=t.text, start=t.idx, end=t.idx + len(t.text), id=i) for i, t in enumerate(sent_doc)],
        )
        new_examples.append(new_example)
    return new_examples
